
/*
 * apricot.c:
 *      A PRogramable Interactive Calculate O Tron...
 *
 *	Copyright (c) 2023 by Gordon Henderson
 *
 *********************************************************************************
 * This file is part of Apricot:
 *    A PRogramable Interactive Calculate O Tron by Gordon Henderson
 *
 *    Apricot is distributed under a "Source Available" license.
 *	It is NOT Open source and must not be treated as such.
 *
 *    See the file LICENSE for details.
 *********************************************************************************
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>

#include "version.h"

#include "tokens.h"
#include "loadSave.h"
#include "readLineS.h"

#include "apricot.h"

ourNum  acc ;
ourNum  variables    [26] ;			// 0-25; A-Z

ourNum *freeRam ;
ourNum *stackData ;
int     freeRamSize, stackSize ;

char    codeStore    [MAX_LINES * MAX_LINE_LEN] ;
int     lineNumbers  [MAX_LINES] ;
ourNum *codePointers [MAX_LINES] ;
char   *codeLines    [MAX_LINES] ;
int     numCodeLines ;
int     pc, pEnd ;
ourNum *stack ;
int     sp, fp ;
ourNum  aux1, aux2 ;
int	programEdited = FALSE ;


/*
 * isVar: getVar: varToAcc:
 *	Check for a variable and retrieve it - copy it to the display register
 *********************************************************************************
 */

static int isVar (char x)
  { return ((x <= 'Z') && (x >= 'A')) ; }

static ourNum getVar (void)
  { return variables [codeStore [pc++] - 'A'] ; }

static void varToAcc (void)
  { acc = getVar () ; }


/*
 * isIndVar: getIndVar: indVarToAcc:
 *	As above for indirect variables.
 *********************************************************************************
 */

static int isIndVar (char x)
  { return (x == TOK_IND) &&  (isVar (codeStore [pc + 1])) ; }

static ourNum getIndVar (void)
{
  ourNum t ;
  int    p ;

  ++pc ;		// Skip over the TOK_IND
  p = variables [codeStore [pc] - 'A'] ;
  if ((p < 0) || (p > freeRamSize))
    { fprintf (stderr, "+++ FreeRAM index out of range: %d\n", p) ; exit (1) ; } // for now
  t = freeRam [p] ;
  ++pc ;
  return t ;
}

static void indVarToAcc (void)
  { acc = getIndVar () ; }


/*
 * allClear:
 *	Zero all variables and the acc.
 *	Except R. We pre-initialise that to a random number.
 *********************************************************************************
 */

void  allClear (void)
{
  int i ;

  acc = 0 ;
  for (i = 0 ; i <= 25 ; ++i)
    variables [i] = 0 ;

  variables ['R' - 'A'] = abs (random ()) ;
}


/*
 * isNum:
 *	See if the token is a number.
 *********************************************************************************
 */

static int isNum (char x)
{
  if ((x <= '9') && (x >= '0'))		// Easy
    return TRUE ;

  if (x != '-')				// Almost easy - not a number
    return FALSE ;

  x = codeStore [pc + 1] ; 		// Character after the minus symbol...
  return (x <= '9') && (x >= '0') ;
}


/*
 * get2810: getBinNum: getDecNum: getOctNum:
 *	Common code to read in base 2, 8 or 10 numbers
 *********************************************************************************
 */

static ourNum get2810 (int base, char end)
{
  ourNum num = 0 ;
  int digit ;

  for (;;)
  {
    digit = codeStore [pc] ;

    if ( (digit <= end) && (digit >= '0') )
    {
      num = num * base + digit - '0' ;
      pc  = pc + 1 ;			// Skip to next digit
      continue ;
    }
    break ;				// Not a valid digit
  }

  return num ;
}

static ourNum getDecNum (void) { return get2810 (10, '9') ; }
static ourNum getOctNum (void) { return get2810 ( 8, '7') ; }
static ourNum getBinNum (void)
{
  pc = pc + 1 ;			// Skip the 'b'
  return get2810 ( 2, '1') ;
}

/*
 * getHexNum:
 *	Read in a hexadecimal number - base 16.
 *********************************************************************************
 */

static ourNum getHexNum (void) 
{
  ourNum num = 0 ;
  int digit ;

  for (;;)
  {
    ++pc ;				// Skip to next digit
    digit = codeStore [pc] ;

    if ((digit >= '0') && (digit <= '9'))
    {
      num = num * 16 + digit - '0' ;
      continue ;
    }

    if ((digit >= 'A') && (digit <= 'F'))
    {
      num = num * 16 + digit - 'A' + 10 ;
      continue ;
    }

    break ;				// Not a hex digit
  }

  return num ;
}


/*
 * getNum:
 *	Return the number (signed integer) at PC
 *********************************************************************************
 */

static ourNum getNum (void)
{
  int neg = FALSE ;
  ourNum num, x ;

// Negative?

  if (codeStore [pc] == '-')
  {
    neg = TRUE ;
    ++pc ;
  }

// Special case for numbers starting with zero

  if (codeStore [pc] == '0')
  {
    ++pc ;				// Skip over that zero
    x  = codeStore [pc] ;
    /**/ if (x == 'X')			//      Hex
      num = getHexNum () ;
    else if (x == 'B')			//      Bin
      num = getBinNum () ;
    else
      num = getOctNum () ;		//	Oct - Also gets single digit zero.
  }
  else
    num = getDecNum () ;		//	Decimal

  return neg ? -num : num ;
}


/*
 * numToAcc:
 *	Read number into the Acc
 *********************************************************************************
 */

static void numToAcc (void)
  { acc = getNum () ; }


/*
 * getAux:
 *	Return the number or variable after the current PC
 *	Used in arithmetic & logic operations
 *********************************************************************************
 */

static ourNum getAux (void)
{
  ++pc ;

  if (isNum    (codeStore [pc])) return getNum    () ;
  if (isVar    (codeStore [pc])) return getVar    () ;
  if (isIndVar (codeStore [pc])) return getIndVar () ;

  return 0 ;
}


/*
 * getAux2:
 *      Return two numbers or variable after the current PC
 *      Used in the muldiv operation.
 *********************************************************************************
 */

static void getAux2 (void)
{
  aux1 = getAux () ;

  if (isNum    (codeStore [pc])) { aux2 = getNum    () ; return ; }
  if (isVar    (codeStore [pc])) { aux2 = getVar    () ; return ; }
  if (isIndVar (codeStore [pc])) { aux2 = getIndVar () ; return ; }
}


/*
 * readNumber:
 *	Read in a line of text and translate to a number
 *********************************************************************************
 */

static ourNum readNumber (void)
{
  char line [40] ;

  for (;;)
  {
    readLineS (line, 10) ;

    if (strlen (line) != 0)
      return atoi (line) ;

    puts ("Number exected.\n+++ Redo from start: ") ;
  }
}


/*
 * dumpVariables:
 *	Print the display register and all 26 variables
 *********************************************************************************
 */

static void dumpVariables (void)
{
  int i ;

  printf ("Display Register: %d", acc) ;
  for (i = 0 ; i < 26 ; ++i)
  {
    if ((i % 4) == 0)
      printf ("\n| ") ;
    printf ("%c: ", i + 'A') ;
    printf ("%11d | ", variables [i]) ;
  }
  printf ("\n") ;
}



/*
 * compileLine:
 *	Compiles "line" into codeStore.
 *
 *	Compiles into codeStore starting at pEnd and updates pEnd
 *	accordingly.
 *
 *	line is a C string also zero terminated.
 *
 *	Can/Will also do error checking at some point.
 *********************************************************************************
 */

int compileLine (char *line)
{
  int from = 0 ;			// "From" pointer - it's a C String

  while (from < (int)strlen (line))
  {
    char c = line [from] ;

    if (c == TOK_STR)			// Start of a string
    {
      do {
        codeStore [pEnd] = c ;
        ++from ;
        ++pEnd ;
        c = line [from] ;
      } while ( (c != TOK_STR) && (c != 0)) ;

      if (c == 0)
        { printf ("+++ Unterminated string\n") ; return FALSE ; }

      codeStore [pEnd] = c ;		// Store terminating TOK_STR
      ++from ;				//   skips over the temrinating TOK_STR
      ++pEnd ;
      continue ;
    }

    if (c == TOK_REM)			// Start of a comment ...
      break ;				// ... and that's it - comments go to the end of the line

    if (isspace (c))			// Ignore all spaces
      { ++from ; continue ; }

    if ( (c >= 'a') && (c <= 'z') )	// Make upper case
      c -= 32 ;

    codeStore [pEnd] = c ;
    ++from ;
    ++pEnd ;
  }

  codeStore [pEnd] = 0 ;

  return TRUE ;
}

static void dumpCode (void)
{
  int i ;
  for (i = 0 ; i < pEnd ; ++i)
    putchar (codeStore [i]) ;
  putchar ('\n') ;
  if (codeStore [i] != 0)
    printf ("No terminating zero\n") ;
}


static int compileProgram (int stats)
{
  int i ;
//  LET tStart, tEnd = ?, ?

  if (numCodeLines == 0)
    { printf ("+++ No program to compile/run\n") ; return FALSE ; }

//  tStart := sys (Sys_cputime)

  pEnd = 0 ;

  for (i = 0 ; i < numCodeLines ; ++i)
     if (!compileLine (codeLines [i]))
      return FALSE ;

  if (stats)
    dumpCode () ;

/*
  tEnd := sys (Sys_cputime)

  IF stats THEN
    writef ("Time taken: %6.3d*n", tEnd - tStart)
*/

  return TRUE ;
}

static void pushPc (void)
{
  stack [fp + sp++] = pc ;
}

/**
static void popPc (void)
{
  pc = stack [fp + --sp] ;
}
**/


/*
 * skipToFinish:
 *	Skps to the current TOK_FINISH (or ELSE)
 *	Has to cater for nested sections.
 *********************************************************************************
 */

static int skipToFinish (int orElse)
{
  int count = 0 ;
  int token ;

  for (;;)
  {
    token = codeStore [pc] ;

    if (token == 0)
    {
      printf ("+++Out of code error\n") ;
      return FALSE ;
    }

    ++pc ;

    if (orElse && (token == TOK_ELSE) && (count == 0) )
	return TRUE ;

    if (token == TOK_FINISH)
    {
      if (count == 0) 
      {
	--sp ;			// un-pop the stack
	return TRUE ;
      }
      else
      {
	--count ;
	continue ;
      }
    }

    if (token == TOK_START)
      ++count ;
  }

  return TRUE ;
}



/*
 * runToken:
 *	Execute a single token/command at the pc
 *	Token is passed in, pc /still/ points to the token.
 *	Exit by leaving the pc pointing to the next instruction/token.
 *********************************************************************************
 */

static int runToken (char token)
{
  int sub, p, aux ;

  switch (token)
  {
    case TOK_STR:
      ++pc ;				// Skip over the TOK_STR token
      while (codeStore [pc] != TOK_STR)
	putchar (codeStore [pc++]) ;
      ++pc ; 				// Skip the terminating TOK_STR
      return TRUE ;

    case TOK_VDU:
      putchar (codeStore [pc++]) ;
      return TRUE ;

    case TOK_PRINT:
      ++pc ;				// Skip over TOK_PRINT
      printf ("%11d", acc) ;
      return TRUE ;

    case TOK_NEWL:
      ++pc ;				// Skip over TOK_NEWL
      printf ("\n") ;
      return TRUE ;

// Arithmetic:
//	Need to get the next token which should be a variable or
//	constant to do the operation on Acc with.

    case TOK_ADD:	acc += getAux () ;	break ;
    case TOK_SUB:	acc -= getAux () ;	break ;
    case TOK_MUL:	acc *= getAux () ;	break ;

    case TOK_DIV:
    case TOK_MOD:
      if ((aux = getAux ()) == 0)
      {
	printf ("+++ divide by zero error\n") ;
	return FALSE ;
      }
      if (token == TOK_DIV)
	acc /= aux ;
      else
	acc %= aux ;
      return TRUE ;

// Logicals - same again

    case TOK_AND:	acc &= getAux () ;	break ;
    case TOK_OR:	acc |= getAux () ;	break ;
    case TOK_EOR:	acc ^= getAux () ;	break ;

// Tests - same again

    case TOK_LT:	acc = acc < getAux () ? 1 : 0 ;	break ;
    case TOK_GT:	acc = acc > getAux () ? 1 : 0 ;	break ;

// MULDIV - A bit special as it needs 2 arguments...

    case TOK_MULDIV:
      if (acc == 0)
      {
	printf ("+++ divide by zero error\n") ;
	return FALSE ;
      }
      getAux2 () ;
      acc = ((int64_t)aux1 * (int64_t)aux2) / acc ;
      break ;

// Assign - it's the same as arithmetic, but needs the next
//	token to be a variable only.

    case TOK_ASSIGN:
      ++pc ;				// Skip over the TOK_ASSIGN
      if (isVar (codeStore [pc]))
      {
	variables [codeStore [pc++] - 'A'] = acc ;
	return TRUE ;
      }
      if (isIndVar (codeStore [pc]))
      {
	++pc ;						// Skip the TOK_IND
	p = variables [codeStore [pc++] - 'A'] ;	// Get index & skip variable
	if ((p < 0) || (p > freeRamSize))
	  { fprintf (stderr, "+++ FreeRAM index out of range: %d\n", p) ; return FALSE ; }
	freeRam [p] = acc ;
	return TRUE ;
      }
      
      printf ("+++ Unknown token after = <%c: %d>\n", codeStore [pc], codeStore [pc]) ;
      return FALSE ;

// Input

    case TOK_INPUT:
      ++pc ;				// skip over the TOK_INPUT
      acc = readNumber () ;
      return TRUE ;

// Section Start

    case TOK_START:
      ++pc ;					// Skip the TOK_START
      pushPc () ;

      if (acc != COT_FALSE)			//	If TRUE carry on, else ...
	return TRUE ;

      return skipToFinish (TRUE) ;

// Section Finish

    case TOK_FINISH:
      if (acc == COT_FALSE)			// False, so break out rather than repeat
      {
	++pc ;					// Skip the FINISH token
	--sp ;					// un-pop the stack
	return TRUE ;
      }

      pc = stack [fp + sp - 1] ;
      return TRUE ;

// Section ELSE
//	Skip forwards to the section finish
//	... because we've just executed the left-hand side, or TRUE side.

    case TOK_ELSE:
      return skipToFinish (FALSE) ;

// Section BREAK:
//	Break out if TRUE

    case TOK_BREAK:
      if (acc == COT_FALSE)			// Don't break, carry on
      {
	++pc ;
	return TRUE ;
      }
      else
	return skipToFinish (FALSE) ;

// Section RESTART: (continue)
//	Re-start if TRUE

    case TOK_RESTART:
      if (acc == COT_FALSE)			// Don't restart, carry on
	++pc ;					// ... skip over the TOK_RESTART
      else
	pc = stack [fp + sp - 1] ;		// ... reset PC to start of section
      return TRUE ;

// Push/Pull

    case TOK_PUSH:
      stack [fp + sp] = acc ;
      ++sp ;
      ++pc ;					// Skip over the TOK_PUSH
      return TRUE ;

    case TOK_PULL:
      --sp ;
      ++pc ;					// Skip over the TOK_PULL
      acc = stack [fp + sp] ;
      return TRUE ;
    
// Call subroutine
//	Create a new stack for the subroutine
//	Need to store the old stack base address and the old stack pointer

    case TOK_CALL:
      ++pc ;					// Skip over TOK_CAL
      sub = codeStore [pc] ;			// Next token is subroutine name: A-Z
      if (!isVar (sub))				// recycle isVar even though it's not really a variable
        { printf ("+++ Bad subroutine name <%c>\n", sub) ; return FALSE ; }

      ++pc ;					// The return address
      p  = 0 ;					// Start at the start - sub optimal?
      while (codeStore [p] != 0)		// Until we run off the end...
      {
        if ((codeStore [p] == TOK_DEF_SUB) && (codeStore [p+1] == sub))	// found it
        {
          stack [fp + sp+0] = fp ;		// Save old stack frame at stack!0 in the new frame
          stack [fp + sp+1] = sp ;			// ... and stack pointer
          stack [fp + sp+2] = pc ;			// ... and return address

	  fp += sp ;
          if (fp > stackSize)
            { printf ("+++ Stack overflow\n") ; return FALSE ; }

          sp       = 3 ;			// Offset 3 as we have data in 0,1 and 2...
          pc       = p + 2 ;			// New PC is after the (x
          return TRUE ;
        }
        ++p ;
      }
      printf ("+++ Subroutine not found\n") ;
      return FALSE ;

// Return from subroutine

    case TOK_END_SUB:
      pc    = stack [fp + 2] ;
      sp    = stack [fp + 1] ;
      fp    = stack [fp + 0] ;
      return TRUE ;
  }
  return TRUE ;
}


static void runProgram (int initialPc)
{
  int  oldPc ;
  char token ;

  pc = initialPc ;

// Stack:
//	stack!0 is previous stack frame pointer
//	stack!1 is previous stack pointer
//	stack!2 is previous PC.

  stack     = stackData ;
  stack [0] = -1 ;
  stack [1] = -1 ;
  stack [2] = -1 ;
  fp        =  0 ;		// Stack frame
  sp        =  3 ;

  for (;;)
  {

// Return silently if we go off the end - probably in immediate mode

    if (pc >= pEnd)
      break ;

    oldPc = pc ;
    token = codeStore [pc] ;

    if (token == TOK_EOP)
    {
      printf ("+++ Program END\n") ;
      break ;
    }

    if (!runToken (token))
      break ;

    if (sp < 3)
    {
      printf ("+++ Stack underflow (%d)\n", sp) ;
      break ;
    }

// We know we ran something because the program counter will have updated, so ...

    if (pc != oldPc)
      continue ;

// ... default to looking for variables or numbers to move into the Acc

    if (isNum    (token)) { numToAcc    () ; continue ; }
    if (isVar    (token)) { varToAcc    () ; continue ; }
    if (isIndVar (token)) { indVarToAcc () ; continue ; }

    printf ("+++ Unknown token <%c> (%d)\n", token, token) ;
    break ;
  }

  if (sp != 3)
    printf ("+++ Stack imbalance sp: %d\n", sp) ;

}


/*
 * printProgram:
 *	Just do it ...
 *********************************************************************************
 */

static void printProgram (void)
{
  int i ;

  for (i = 0 ; i <  numCodeLines ; ++i)
    printf ("%4d %s\n", lineNumbers [i], codeLines [i]) ;
  printf ("\n") ;
}


/*
 * getLineIndex:
 *	Return the index in the lineNumbers array of the supplied line number.
 *	(or -1)
 *********************************************************************************
 */

static int getLineIndex (int lineNum)
{
  int i ;

  for (i = 0 ; i < MAX_LINES ; ++i)
    if (lineNumbers [i] == lineNum)
      return i ;

  return -1 ;
}


/*
 * deleteLine:
 *	Delete a line from the program store
 *********************************************************************************
 */

static void deleteLine (int lineNum)
{
  int i ;
  int idx = getLineIndex (lineNum) ;

  if (idx == -1)
    { printf ("Line %d does not exist.\n", lineNum) ; return ; }

  if (codeLines [idx] == NULL)
  {
    printf ("Delete line: Error - something is wrong on line %d at index: %d\n", lineNum, idx) ;
    return ;
  }

  free (codeLines [idx]) ;

// Move all lines below up one

  --numCodeLines ;

  for (i = idx  ; i < numCodeLines ; ++i)
  {
    int next = i + 1 ;
    lineNumbers [i] = lineNumbers [next] ;
    codeLines   [i] = codeLines   [next] ;
  }

// ... and zero-out the last entry

  lineNumbers [numCodeLines] = 0 ;
  codeLines   [numCodeLines] = NULL ;
}

static void insertNewLine (int lineNum, char *line, int p)
{
  int idx = 0 ;
  int i ;

// Search the line numbers until we find one greater than the one we're inserting...

  for (i = 0 ; i <= numCodeLines ; ++i)
  {
    int lNum = lineNumbers [i] ;
    if ((lNum == 0) || (lNum > lineNum))
      { idx = i ; break ; }
  }

// Move everything down a location

  for (i = numCodeLines - 1 ; i >= idx ; --i)
  {
    int moveTo = i + 1 ;
    lineNumbers [moveTo] = lineNumbers [i] ;
    codeLines   [moveTo] = codeLines   [i] ;
  }

// Insert the line at idx place

  lineNumbers [idx] = lineNum ;
  codeLines   [idx] = malloc (MAX_LINE_LEN + 1) ;
  strcpy (codeLines [idx], line + p) ;
  ++numCodeLines ;
}


/*
 * insertLine:
 *	Insert a line with line number into the program
 *********************************************************************************
 */

void insertLine (int lineNum, char *line, int p)
{
  int idx = getLineIndex (lineNum) ;

  if (idx == -1)				// No existing line
    insertNewLine (lineNum, line, p) ;
  else
  {

// We don't need to free/malloc the line vector as they're all the same length, so
//      we can save any potential heap fragmentation by just re-using it...

    strcpy (codeLines [idx], line + p) ;
  }
}


/*
 * handleLineNumber:
 *      Deal with a line that starts with a line number
 *      Basically, store it away...
 *********************************************************************************
 */

static void handleLineNumer (char *line)
{
  int lineNum = 0 ;
  int len = strlen (line) ;
  int p = 0 ;

  while (p < len)
  {
    int c = line [p] ;
    if ((c > '9') || (c < '0'))				// Not a digit
      break ;

    lineNum = lineNum * 10 + c - '0' ;
    if (lineNum > MAX_LINE_NUM)
      { printf ("+++ Line number is too big.\n") ; return ; }
    ++p ;
  }

  if (lineNum == 0)
    { printf ("Can't have line number 0, that's mine.\n") ; return ; }

// Check next character - if end of the line then we delete, else it must be a space

  if (p == len)
  {
    deleteLine (lineNum) ;
    return ;
  }

  if (line [p] != ' ')
    { printf ("Missing SPACE after the line number\n") ; return ; }

// OK - We have a line number and the rest of the line ...

  insertLine (lineNum, line, p+1) ;
}


/*
 * renumber:
 *	Very simple renumbering process - from 100 by 10.
 *********************************************************************************
 */

static void renumber (void)
{
  int i ;
  int start = 100 ;

  if (numCodeLines == 0)
    { printf ("+++ No program to renumber\n") ; return ; }

  for (i = 0 ; i < numCodeLines ; ++i)
  {
    lineNumbers [i] = start ;
    start += 10 ;
  }
}



static void editHelp (void)
{
  printf ("Commands: \n") ;
  printf ("  I: Back to Interactive Mode\n") ;
  printf ("  P: Print Program\n") ;
  printf ("  R: Run\n") ;
  printf ("  N: ReNumber\n") ;
  printf ("  E: Erase Program\n") ;
  printf ("  V: Display Variables\n") ;
  printf ("  L: Load     |  S: Save\n") ;
}


/*
 * editorMode:
 *      Run Editor Mode to allow us to type in code and have it stored
 *      for later execution.
 *********************************************************************************
*/

static void editorMode (void)
{
  char lineBuffer [80] ;			// 80 character lines
  int c ;

  printf ("Edit Mode. Type ? For Help\n\n") ;

  for (;;)
  {
    printf ("> ") ;
    readLineS (lineBuffer, 76) ;

    if (strlen (lineBuffer) == 0)
      continue ;

    c = lineBuffer [0] ;			// Check first char for number

// Check for a line number

    if ((c <= '9') && (c >= '0'))
    {
      handleLineNumer (lineBuffer) ;
      continue ;
    }

    //IF lineBuffer%1 = '**' THEN { osCli (lineBuffer) ; LOOP }

// Check command - single characters

    if (strlen (lineBuffer) != 1)
      { printf ("+++ Commands are just one character\n") ; continue ; }

    if ((c >= 'a') && (c <= 'z'))		// Convert to upper case
      c -= 32 ;

    if (c == 'I')
    {
      printf ("Back to Interactive Mode\n") ;
      return ;
    }

    if (c == 'R')
    {
      if (compileProgram (FALSE))
        runProgram (0) ;
      continue ;
    }

    if (c == 'L') { loadProgram    ()                 ; continue ; }
    if (c == 'S') { saveProgram    ()                 ; continue ; }
    if (c == 'E') { eraseProgram   (FALSE)            ; continue ; }
    if (c == 'P') { printProgram   ()                 ; continue ; }
    if (c == 'N') { renumber       ()                 ; continue ; }
    if (c == 'V') { dumpVariables  ()                 ; continue ; }
    if (c == 'C') { compileProgram (TRUE)             ; continue ; }
    if (c == '?') { editHelp       ()                 ; continue ; }

    printf ("+++ Unknown command\n") ;

  }
}


/*
 * interactiveHelp:
 *	Provide a list of commands in interactive mode
 *********************************************************************************
 */

void interactiveHelp (void)
{
  printf ("Interactive mode commands: \n") ;
  printf ("  Q: Quit\n") ;
  printf ("  E: Enter Editor Mode.\n") ;
  printf ("  P: Print Program.\n") ;
  printf ("  L: Load Program.\n") ;
  printf ("  R: RuRun Program.\n") ;
  printf ("  C: Clear Variables and display register.\n") ;
  printf ("  V: Display Variables\n\n") ;
}


/*
 * processCommand:
 *	Handle a command while in Interactive Mode
 *********************************************************************************
 */

int processCommand (char *line)
{
  int command = line [1] ;		// [0] is the colon/command prefix

  if ( ('a' <= command) && (command <= 'z'))
    command -= 32 ;			// Make upper case

  switch (command)
  {
    case '?':
      interactiveHelp () ;
      break ;

    case 'V':				// Dump all variables
      dumpVariables () ;
      break ;

    case 'C':				// All Clear
      allClear () ;
      printf ("+++ Clear\n") ;
      break ;

    case 'E':
      editorMode () ;
      break ;

    case 'P':
      printProgram () ;
      break ;

    case 'L':
      loadProgram () ;
      break ;

    case 'R':
      runProgram (0) ;
      break ;

    case 'Q':
      printf ("+++ Ended\n") ;
      return FALSE ;

    default:
      printf ("+++ Unknown command\n") ;
  }

  return TRUE ;
}


/*
 * interactiveMode:
 *	Run the interactive calculate o tron
 *********************************************************************************
 */

void interactiveMode (void)
{
  char lineBuffer [80] ;
  int  ourStart ;

  printf ("Ready\n") ;

  for (;;)
  {
    printf ("%11d: ", acc) ;
    readLineS (lineBuffer, 65) ;

// Command?

    if (*lineBuffer == TOK_CMD)
    {
      if (processCommand (lineBuffer))
	continue ;
      else
        return ;
    }

    ourStart = pEnd ;
    if (!compileLine (lineBuffer))
      continue ;

    runProgram (ourStart) ;
    pEnd = ourStart ;

  }
}


/*
 * allocateData:
 *	Wrapper round malloc just in-case it fails
 *********************************************************************************
 */

char *allocateData (int size)
{
  char *r = malloc (size) ;
  if (r == 0)
    { printf ("+++ Out of memory\n") ; exit (1) ; }

  return r ;
}


/*
 * main
 *	Is where it's at...
 *********************************************************************************
 */

int main (int argc, char *argv [])
{
  int i ;
  int opt ;
  int goEditMode = FALSE ;

  freeRamSize = DEFAULT_MAX_FREE_RAM ;
  stackSize   = DEFAULT_MAX_STACK ;

  while ((opt = getopt (argc, argv, "es:m:")) != -1)
  {
    switch (opt)
    {
      case 's':
        stackSize = atoi (optarg) ;
	if (stackSize < 16)
	  { fprintf (stderr, "+++ Stack too small\n") ; exit (1) ; }
	break ;

      case 'm':
        freeRamSize = atoi (optarg) ;
	if (freeRamSize < 16)
	  { fprintf (stderr, "+++ Free RAM too small\n") ; exit (1) ; }
	break ;

      case 'e':
	goEditMode = TRUE ;
	break ;

      default:
	fprintf (stderr, "+++ Usage: apricot [-s xxx] [-m xxx]\n") ;
	exit (1) ;
    }
  }

  if ( (freeRam = (ourNum *)malloc (freeRamSize * sizeof (ourNum))) == NULL)
    { fprintf (stderr, "+++ Insufficient memory for free RAM\n") ; exit (1) ; }

  if ((stackData = malloc (stackSize * sizeof (int))) == NULL)
    { fprintf (stderr, "+++ Insufficient memory for stack\n") ; exit (1) ; }


  for (i = 0 ; i < MAX_LINES ; ++i)
  {
    lineNumbers  [i] = 0 ;
    codePointers [i] = 0 ;
    codeLines    [i] = NULL ;
  }

  pEnd         = 0 ;
  numCodeLines = 0 ;
  sp = fp      = 0 ;

  allClear () ;

  printf ("\n\n") ;
  printf ("APRICOT: A PRogrammable Interactive Calculate O Tron.\n") ;
  printf ("  v%s. Copyright (c) 2023 by Gordon Henderson.\n", VERSION) ;
  printf ("  Up to %d numbers of free store available.\n\n", freeRamSize) ;

  if (goEditMode)
    editorMode () ;
  else
    interactiveMode () ;

  return 0 ;
}
